/*
 * $Header: /home/cvspublic/jakarta-commons-sandbox/resources/src/java/org/apache/commons/resources/impl/ResourceBundleResources.java,v 1.5 2003/01/05 03:29:57 craigmcc Exp $
 * $Revision: 1.5 $
 * $Date: 2003/01/05 03:29:57 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999-2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Commons", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */


package org.apache.commons.resources.impl;


import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.TimeZone;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.resources.Resources;
import org.apache.commons.resources.ResourcesException;
import org.apache.commons.resources.ResourcesKeyException;


/**
 * <p>Concrete implementation of {@link Resources} that wraps a set
 * (one per Locale) of <code>java.util.ResourceBundle</code> instances
 * that share a common base name.  The <code>timeZone</code> argument
 * is ignored in all resource getter method implementations.</p>
 *
 * @author Craig R. McClanahan
 * @version $Revision: 1.5 $ $Date: 2003/01/05 03:29:57 $
 */

public class ResourceBundleResources extends ResourcesBase {


    // ----------------------------------------------------------- Constructors


    /**
     * <p>Create a new {@link Resources} instance with the specified
     * logical name and bundle base name.</p>
     *
     * @param name Logical name of the new instance
     * @param base Fully qualified base name of the <code>ResourceBundle</code>
     *  instances to be wrapped
     */
    public ResourceBundleResources(String name, String base) {

        super(name);
        this.base = base;

    }


    // ----------------------------------------------------- Instance Variables


    /**
     * <p>The fully qualified base name of the <code>ResourceBundle</code>
     * instances to be wrapped.</p>
     */
    protected String base = null;


    /**
     * <p>Two-level cache of <code>ResourceBundle</code> instances
     * that have previously been acquired by this {@link Resources}
     * instance.  The first-level key is the <code>ClassLoader</code>
     * instance used to load the bundles.  The first-level value is a
     * <code>java.util.Map</code> that is then, in turn, keyed by the
     * <code>java.util.Locale</code> to which a particular bundle belongs.</p>
     */
    protected Map bundles = new HashMap();


    /**
     * <p>The logging instance for this class.</p>
     */
    protected static Log log =
        LogFactory.getLog(ResourceBundleResources.class);


    // ------------------------------------------------------ Lifecycle Methods


    /**
     * <p>This must be called to initialize the data content of this
     * {@link Resources} instance, before any of the <code>getXxx()</code>
     * methods are called.</p>
     *
     * @exception ResourcesException if an error occurs during initialization
     */
    public void init() throws ResourcesException {

        if (log.isDebugEnabled()) {
            log.debug("Initializing for base name '" + base + "'");
        }

    }


    /**
     * <p>This method must be called when the manager of this resource
     * decides that it's no longer needed.  After this method is called,
     * no further calls to any of the <code>getXxx()</code> methods are
     * allowed.</p>
     *
     * @exception ResourcesException if an error occurs during finalization
     */
    public void destroy() throws ResourcesException {

        if (log.isDebugEnabled()) {
            log.debug("Finalizing for base name '" + base + "'");
        }
        synchronized (bundles) {
            Iterator loaders = bundles.keySet().iterator();
            while (loaders.hasNext()) {
                ClassLoader loader = (ClassLoader) loaders.next();
                ((Map) bundles.get(loader)).clear();
            }
            bundles.clear();
        }

    }


    // ------------------------------------------------------------- Properties


    /**
     * <p>Return the fully qualified base name of the
     * <code>ResourceBundle</code> instances we are wrapping.</p>
     */
    public String getBase() {

        return (this.base);

    }


    // ---------------------------------------------- Content Retrieval Methods


    /**
     * <p>Return the content for the specified <code>key</code> as a
     * byte array, localized based on the specified <code>locale</code>
     * and/or <code>timeZone</code>.</p>
     *
     * @param key Identifier for the requested content
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception ResourcesException if an error occurs retrieving or
     *  returning the requested content
     * @exception ResourcesKeyException if the no value for the specified
     *  key was found, and <code>isReturnNull()</code> returns
     *  <code>false</code>
     */
    public byte[] getBytes(String key, Locale locale, TimeZone timeZone)
        throws ResourcesException {

        throw new UnsupportedOperationException(); // FIXME

    }


    /**
     * <p>Return the content for the specified <code>key</code> as an
     * InputStream, localized based on the specified <code>locale</code>
     * and/or <code>timeZone</code>.</p>
     *
     * @param key Identifier for the requested content
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception ResourcesException if an error occurs retrieving or
     *  returning the requested content
     * @exception ResourcesKeyException if the no value for the specified
     *  key was found, and <code>isReturnNull()</code> returns
     *  <code>false</code>
     */
    public InputStream getInputStream(String key, Locale locale,
                                      TimeZone timeZone) {

        throw new UnsupportedOperationException();  // FIXME

    }


    /**
     * <p>Return the content for the specified <code>key</code> as an
     * Object, localized based on the specified <code>locale</code>
     * and/or <code>timeZone</code>.</p>
     *
     * @param key Identifier for the requested content
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception ResourcesException if an error occurs retrieving or
     *  returning the requested content
     * @exception ResourcesKeyException if the no value for the specified
     *  key was found, and <code>isReturnNull()</code> returns
     *  <code>false</code>
     */
    public Object getObject(String key, Locale locale, TimeZone timeZone) {

        try {
            ResourceBundle bundle = getBundle(locale, timeZone);
            return (bundle.getObject(key));
        } catch (MissingResourceException e) {
            if (isReturnNull()) {
                return (null);
            } else {
                throw new ResourcesKeyException(key);
            }
        }

    }


    /**
     * <p>Return the content for the specified <code>key</code> as a
     * Reader, localized based on the specified <code>locale</code>
     * and/or <code>timeZone</code>.</p>
     *
     * @param key Identifier for the requested content
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception ResourcesException if an error occurs retrieving or
     *  returning the requested content
     * @exception ResourcesKeyException if the no value for the specified
     *  key was found, and <code>isReturnNull()</code> returns
     *  <code>false</code>
     */
    public Reader getReader(String key, Locale locale, TimeZone timeZone) {

        throw new UnsupportedOperationException(); // FIXME

    }


    /**
     * <p>Return the content for the specified <code>key</code> as a
     * String, localized based on the specified <code>locale</code>
     * and/or <code>timeZone</code>.</p>
     *
     * @param key Identifier for the requested content
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception ResourcesException if an error occurs retrieving or
     *  returning the requested content
     * @exception ResourcesKeyException if the no value for the specified
     *  key was found, and <code>isReturnNull()</code> returns
     *  <code>false</code>
     */
    public String getString(String key, Locale locale, TimeZone timeZone) {

        if (log.isTraceEnabled()) {
            log.trace("Retrieving message for key '" + key + "' for locale '"
                      + locale + "'");
        }
        try {
            ResourceBundle bundle = getBundle(locale, timeZone);
            String message = bundle.getString(key);
            if (log.isTraceEnabled()) {
                log.trace("Retrieved message is '" + message + "'");
            }
            return (message);
        } catch (ClassCastException e) {
            throw new ResourcesException(e);
        } catch (MissingResourceException e) {
            if (log.isTraceEnabled()) {
                log.trace("No message found");
            }
            if (isReturnNull()) {
                return (null);
            } else {
                throw new ResourcesKeyException(key);
            }
        }

    }



    // ------------------------------------------------------ Protected Methods


    /**
     * <p>Return the appropriate <code>ResourceBundle</code> instance
     * that corresponds to the specified <code>locale</code> and
     * <code>timeZone</code> parameters.  The first time a particular
     * bundle is requested, cache it so that subsequent requests will
     * operate more quickly.</p>
     *
     * <p>The default implementation of this method unconditionally ignores the
     * <code>timeZone</code> argument.</p>
     *
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception MissingResourceException if the requested Resourcebundle
     *  cannot be acquired
     */
    protected ResourceBundle getBundle(Locale locale, TimeZone timeZone)
        throws MissingResourceException {

        if (locale == null) {
            locale = Locale.getDefault();
        }

        synchronized (bundles) {

            // Locate or create the Map for the appropriate ClassLoader
            ClassLoader loader =
                Thread.currentThread().getContextClassLoader();
            if (loader == null) {
                loader = this.getClass().getClassLoader();
            }
            Map loaderMap = (Map) bundles.get(loader);
            if (loaderMap == null) {
                loaderMap = new HashMap();
                bundles.put(loader, loaderMap);
            }

            // Locate or create the requested ResourceBundle instance
            ResourceBundle bundle =
                (ResourceBundle) loaderMap.get(locale);
            if (bundle == null) {
                bundle = ResourceBundle.getBundle(base, locale, loader);
                loaderMap.put(locale, bundle);
            }
            return (bundle);

        }

    }


}
